from . import namespace  # noqa
from . import rbac  # noqa
from . import webhook  # noqa
from . import service  # noqa
from . import deployment  # noqa

COMPONENTS = [namespace, rbac, webhook, deployment, service]
